@@ -36,131 +36,32 @@
 #endif
 
 /************************************************************************
- *									*
- *			Module interfaces				*
- *									*
+ * 									*
+ * 		Hooks for the document loader				*
+ * 									*
  ************************************************************************/
 
 /**
- * xsltNewDocument:
- * @ctxt: an XSLT transformation context (or NULL)
- * @doc:  a parsed XML document
+ * xsltDocDefaultLoaderFunc:
+ * @URI: the URI of the document to load
+ * @dict: the dictionnary to use when parsing that document
+ * @options: parsing options, a set of xmlParserOption
+ * @ctxt: the context, either a stylesheet or a transformation context
+ * @type: the xsltLoadType indicating the kind of loading required
  *
- * Register a new document, apply key computations
+ * Default function to load document not provided by the compilation or
+ * transformation API themselve, for example when an xsl:import,
+ * xsl:include is found at compilation time or when a document()
+ * call is made at runtime.
  *
- * Returns a handler to the document
- */
-xsltDocumentPtr	
-xsltNewDocument(xsltTransformContextPtr ctxt, xmlDocPtr doc) {
-    xsltDocumentPtr cur;
-
-    cur = (xsltDocumentPtr) xmlMalloc(sizeof(xsltDocument));
-    if (cur == NULL) {
-	xsltTransformError(ctxt, NULL, (xmlNodePtr) doc,
-		"xsltNewDocument : malloc failed\n");
-	return(NULL);
-    }
-    memset(cur, 0, sizeof(xsltDocument));
-    cur->doc = doc;
-    if (ctxt != NULL) {
-	cur->next = ctxt->docList;
-	ctxt->docList = cur;
-	xsltInitCtxtKeys(ctxt, cur);
-    }
-    return(cur);
-}
-
-/**
- * xsltNewStyleDocument:
- * @style: an XSLT style sheet
- * @doc:  a parsed XML document
- *
- * Register a new document, apply key computations
- *
- * Returns a handler to the document
- */
-xsltDocumentPtr	
-xsltNewStyleDocument(xsltStylesheetPtr style, xmlDocPtr doc) {
-    xsltDocumentPtr cur;
-
-    cur = (xsltDocumentPtr) xmlMalloc(sizeof(xsltDocument));
-    if (cur == NULL) {
-	xsltTransformError(NULL, style, (xmlNodePtr) doc,
-		"xsltNewStyleDocument : malloc failed\n");
-	return(NULL);
-    }
-    memset(cur, 0, sizeof(xsltDocument));
-    cur->doc = doc;
-    if (style != NULL) {
-	cur->next = style->docList;
-	style->docList = cur;
-    }
-    return(cur);
-}
-
-/**
- * xsltFreeStyleDocuments:
- * @style: an XSLT style sheet
- *
- * Free up all the space used by the loaded documents
- */
-void	
-xsltFreeStyleDocuments(xsltStylesheetPtr style) {
-    xsltDocumentPtr doc, cur;
-
-    cur = style->docList;
-    while (cur != NULL) {
-	doc = cur;
-	cur = cur->next;
-	xsltFreeDocumentKeys(doc);
-	if (!doc->main)
-	    xmlFreeDoc(doc->doc);
-        xmlFree(doc);
-    }
-}
-
-/**
- * xsltFreeDocuments:
- * @ctxt: an XSLT transformation context
- *
- * Free up all the space used by the loaded documents
- */
-void	
-xsltFreeDocuments(xsltTransformContextPtr ctxt) {
-    xsltDocumentPtr doc, cur;
-
-    cur = ctxt->docList;
-    while (cur != NULL) {
-	doc = cur;
-	cur = cur->next;
-	xsltFreeDocumentKeys(doc);
-	if (!doc->main)
-	    xmlFreeDoc(doc->doc);
-        xmlFree(doc);
-    }
-    cur = ctxt->styleList;
-    while (cur != NULL) {
-	doc = cur;
-	cur = cur->next;
-	xsltFreeDocumentKeys(doc);
-	if (!doc->main)
-	    xmlFreeDoc(doc->doc);
-        xmlFree(doc);
-    }
-}
-
-/**
- * xsltParseDocument:
- * @URI: the URI to load.
- * @dict: the dictionnary to use
- * @options: the options to use
- *
- * Specialized parsing function
- *
- * Returns NULL in case of error or the Document parsed.
+ * Returns the pointer to the document (which will be modified and
+ * freed by the engine later), or NULL in case of error.
  */
 static xmlDocPtr
-xsltParseDocument(const xmlChar *URI, xmlDictPtr dict, int options) {
+xsltDocDefaultLoaderFunc(const xmlChar * URI, xmlDictPtr dict, int options,
+                         void *ctxt ATTRIBUTE_UNUSED,
+			 xsltLoadType type ATTRIBUTE_UNUSED)
+{
     xmlParserCtxtPtr pctxt;
     xmlParserInputPtr inputStream;
     xmlDocPtr doc;
@@ -205,6 +106,176 @@ xsltParseDocument(const xmlChar *URI, xmlDictPtr dict, int options) {
     return(doc);
 }
 
+
+xsltDocLoaderFunc xsltDocDefaultLoader = xsltDocDefaultLoaderFunc;
+
+/**
+ * xsltSetLoaderFunc:
+ * @f: the new function to handle document loading.
+ *
+ * Set the new function to load document, if NULL it resets it to the
+ * default function.
+ */
+ 
+void
+xsltSetLoaderFunc(xsltDocLoaderFunc f) {
+    if (f == NULL)
+        xsltDocDefaultLoader = xsltDocDefaultLoaderFunc;
+    else
+        xsltDocDefaultLoader = f;
+}
+
+/************************************************************************
+ *									*
+ *			Module interfaces				*
+ *									*
+ ************************************************************************/
+
+/**
+ * xsltNewDocument:
+ * @ctxt: an XSLT transformation context (or NULL)
+ * @doc:  a parsed XML document
+ *
+ * Register a new document, apply key computations
+ *
+ * Returns a handler to the document
+ */
+xsltDocumentPtr	
+xsltNewDocument(xsltTransformContextPtr ctxt, xmlDocPtr doc) {
+    xsltDocumentPtr cur;
+
+    cur = (xsltDocumentPtr) xmlMalloc(sizeof(xsltDocument));
+    if (cur == NULL) {
+	xsltTransformError(ctxt, NULL, (xmlNodePtr) doc,
+		"xsltNewDocument : malloc failed\n");
+	return(NULL);
+    }
+    memset(cur, 0, sizeof(xsltDocument));
+    cur->doc = doc;
+    if (ctxt != NULL) {
+        if (! XSLT_IS_RES_TREE_FRAG(doc)) {
+	    cur->next = ctxt->docList;
+	    ctxt->docList = cur;
+	}
+#ifdef XSLT_REFACTORED_KEYCOMP
+	/*
+	* A key with a specific name for a specific document
+	* will only be computed if there's a call to the key()
+	* function using that specific name for that specific
+	* document. I.e. computation of keys will be done in
+	* xsltGetKey() (keys.c) on an on-demand basis.
+	*/
+#else
+	/*
+	* Old behaviour.
+	*/
+	xsltInitCtxtKeys(ctxt, cur);
+#endif
+    }
+    return(cur);
+}
+
+/**
+ * xsltNewStyleDocument:
+ * @style: an XSLT style sheet
+ * @doc:  a parsed XML document
+ *
+ * Register a new document, apply key computations
+ *
+ * Returns a handler to the document
+ */
+xsltDocumentPtr	
+xsltNewStyleDocument(xsltStylesheetPtr style, xmlDocPtr doc) {
+    xsltDocumentPtr cur;
+
+    cur = (xsltDocumentPtr) xmlMalloc(sizeof(xsltDocument));
+    if (cur == NULL) {
+	xsltTransformError(NULL, style, (xmlNodePtr) doc,
+		"xsltNewStyleDocument : malloc failed\n");
+	return(NULL);
+    }
+    memset(cur, 0, sizeof(xsltDocument));
+    cur->doc = doc;
+    if (style != NULL) {
+	cur->next = style->docList;
+	style->docList = cur;
+    }
+    return(cur);
+}
+
+/**
+ * xsltFreeStyleDocuments:
+ * @style: an XSLT stylesheet (representing a stylesheet-level)
+ *
+ * Frees the node-trees (and xsltDocument structures) of all
+ * stylesheet-modules of the stylesheet-level represented by
+ * the given @style. 
+ */
+void	
+xsltFreeStyleDocuments(xsltStylesheetPtr style) {
+    xsltDocumentPtr doc, cur;
+#ifdef XSLT_REFACTORED_XSLT_NSCOMP
+    xsltNsMapPtr nsMap;
+#endif
+    
+    if (style == NULL)
+	return;
+
+#ifdef XSLT_REFACTORED_XSLT_NSCOMP
+    if (XSLT_HAS_INTERNAL_NSMAP(style))
+	nsMap = XSLT_GET_INTERNAL_NSMAP(style);
+    else
+	nsMap = NULL;    
+#endif   
+
+    cur = style->docList;
+    while (cur != NULL) {
+	doc = cur;
+	cur = cur->next;
+#ifdef XSLT_REFACTORED_XSLT_NSCOMP
+	/*
+	* Restore all changed namespace URIs of ns-decls.
+	*/
+	if (nsMap)
+	    xsltRestoreDocumentNamespaces(nsMap, doc->doc);
+#endif
+	xsltFreeDocumentKeys(doc);
+	if (!doc->main)
+	    xmlFreeDoc(doc->doc);
+        xmlFree(doc);
+    }
+}
+
+/**
+ * xsltFreeDocuments:
+ * @ctxt: an XSLT transformation context
+ *
+ * Free up all the space used by the loaded documents
+ */
+void	
+xsltFreeDocuments(xsltTransformContextPtr ctxt) {
+    xsltDocumentPtr doc, cur;
+
+    cur = ctxt->docList;
+    while (cur != NULL) {
+	doc = cur;
+	cur = cur->next;
+	xsltFreeDocumentKeys(doc);
+	if (!doc->main)
+	    xmlFreeDoc(doc->doc);
+        xmlFree(doc);
+    }
+    cur = ctxt->styleList;
+    while (cur != NULL) {
+	doc = cur;
+	cur = cur->next;
+	xsltFreeDocumentKeys(doc);
+	if (!doc->main)
+	    xmlFreeDoc(doc->doc);
+        xmlFree(doc);
+    }
+}
+
 /**
  * xsltLoadDocument:
  * @ctxt: an XSLT transformation context
@@ -249,7 +320,8 @@ xsltLoadDocument(xsltTransformContextPtr ctxt, const xmlChar *URI) {
 	ret = ret->next;
     }
 
-    doc = xsltParseDocument(URI, ctxt->dict, ctxt->parserOptions);
+    doc = xsltDocDefaultLoader(URI, ctxt->dict, ctxt->parserOptions,
+                               (void *) ctxt, XSLT_LOAD_DOCUMENT);
 
     if (doc == NULL)
 	return(NULL);
@@ -324,7 +396,8 @@ xsltLoadStyleDocument(xsltStylesheetPtr style, const xmlChar *URI) {
 	ret = ret->next;
     }
 
-    doc = xsltParseDocument(URI, style->dict, XSLT_PARSE_OPTIONS);
+    doc = xsltDocDefaultLoader(URI, style->dict, XSLT_PARSE_OPTIONS,
+                               (void *) style, XSLT_LOAD_STYLESHEET);
     if (doc == NULL)
 	return(NULL);
 
