@@ -12,11 +12,7 @@
 #ifndef __XML_XSLTUTILS_H__
 #define __XML_XSLTUTILS_H__
 
-#if defined(WIN32) && defined(_MSC_VER)
-#include <libxslt/xsltwin32config.h>
-#else
 #include <libxslt/xsltconfig.h>
-#endif
 #ifdef HAVE_STDARG_H
 #include <stdarg.h>
 #endif
